﻿using System;

namespace TestProject
{
    /// <summary>
    /// 신경망
    /// </summary>
    public class NeuralNetwork
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Static
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 난수 발생기
        /// </summary>
        private static Random _random;

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Instance
        //////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 입력 노드 카운트
        /// </summary>
        private int inputNodeCount;

        /// <summary>
        /// 은닉 노드 카운트
        /// </summary>
        private int hiddenNodeCount;

        /// <summary>
        /// 출력 노드 카운트
        /// </summary>
        private int outputNodeCount;


        /// <summary>
        /// 입력 노드 값 배열
        /// </summary>
        private double[] inputNodeValueArray;

        /// <summary>
        /// 입력-은닉 가중치 배열
        /// </summary>
        private double[][] inputHiddenWeightArray;

        /// <summary>
        /// 은닉 노드 바이어스 배열
        /// </summary>
        private double[] hiddenNodeBiasArray;

        /// <summary>
        /// 은닉 노드 값 배열
        /// </summary>
        private double[] hiddenNodeValueArray;

        /// <summary>
        /// 은닉-출력 가중치 배열
        /// </summary>
        private double[][] hiddenOutputWeightArray;

        /// <summary>
        /// 출력 노드 바이어스 배열
        /// </summary>
        private double[] outputNodeBiasArray;

        /// <summary>
        /// 출력 노드 값 배열
        /// </summary>
        private double[] outputNodeValueArray;


        /// <summary>
        /// 은닉 노드 그라디언트 배열
        /// </summary>
        private double[] hiddenNodeGradientArray;

        /// <summary>
        /// 출력 노드 그라디언트 배열
        /// </summary>
        private double[] outputNodeGradientArray;


        /// <summary>
        /// 입력-은닉 이전 가중치 델타 배열
        /// </summary>
        private double[][] inputHiddenPreviousWeightDeltaArray;

        /// <summary>
        /// 은닉 노드 이전 바이어스 델타 배열
        /// </summary>
        private double[] hiddenNodePreviousBiasDeltaArray;

        /// <summary>
        /// 은닉-출력 이전 가중치 델타 배열
        /// </summary>
        private double[][] hiddenOutputPreviousWeightDeltaArray;

        /// <summary>
        /// 출력 노드 이전 바이어스 델타 배열
        /// </summary>
        private double[] outputNodePreviousBiasDeltaArray;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - NeuralNetwork(inputNodeCount, hiddenNodeCount, outputNodeCount)

        /// <summary>
        /// 생성자
        /// </summary>
        /// <param name="inputNodeCount">입력 노드 카운트</param>
        /// <param name="hiddenNodeCount">은닉 노드 카운트</param>
        /// <param name="outputNodeCount">출력 노드 카운트</param>
        public NeuralNetwork(int inputNodeCount, int hiddenNodeCount, int outputNodeCount)
        {
            _random = new Random(0);
 
            this.inputNodeCount  = inputNodeCount;
            this.hiddenNodeCount = hiddenNodeCount;
            this.outputNodeCount = outputNodeCount;
 
            this.inputNodeValueArray     = new double[inputNodeCount];
            this.inputHiddenWeightArray  = GetMatrixArray(inputNodeCount, hiddenNodeCount);
            this.hiddenNodeBiasArray     = new double[hiddenNodeCount];
            this.hiddenNodeValueArray    = new double[hiddenNodeCount];
            this.hiddenOutputWeightArray = GetMatrixArray(hiddenNodeCount, outputNodeCount);
            this.outputNodeBiasArray     = new double[outputNodeCount];
            this.outputNodeValueArray    = new double[outputNodeCount];
 
            this.InitializeWeightArray();
 
            this.hiddenNodeGradientArray = new double[hiddenNodeCount];
            this.outputNodeGradientArray = new double[outputNodeCount];
 
            this.inputHiddenPreviousWeightDeltaArray  = GetMatrixArray(inputNodeCount, hiddenNodeCount);
            this.hiddenNodePreviousBiasDeltaArray     = new double[hiddenNodeCount];
            this.hiddenOutputPreviousWeightDeltaArray = GetMatrixArray(hiddenNodeCount, outputNodeCount);
            this.outputNodePreviousBiasDeltaArray     = new double[outputNodeCount];
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 가중치 배열 구하기 - GetWeightArray()

        /// <summary>
        /// 가중치 배열 구하기
        /// </summary>
        /// <returns>가중치 배열</returns>
        public double[] GetWeightArray()
        {
            int totalWeightCount = (this.inputNodeCount  * this.hiddenNodeCount) +
                                   (this.hiddenNodeCount * this.outputNodeCount) +
                                    this.hiddenNodeCount + this.outputNodeCount;
 
            double[] targetWeightArray = new double[totalWeightCount];

            int k = 0;

            for(int i = 0; i < this.inputNodeCount; i++)
            {
                for(int j = 0; j < this.hiddenNodeCount; j++)
                {
                    targetWeightArray[k++] = this.inputHiddenWeightArray[i][j];
                }
            }
 
            for(int i = 0; i < this.hiddenNodeCount; i++)
            {
                targetWeightArray[k++] = this.hiddenNodeBiasArray[i];
            }

            for(int i = 0; i < this.hiddenNodeCount; i++)
            {
                for(int j = 0; j < this.outputNodeCount; j++)
                {
                    targetWeightArray[k++] = this.hiddenOutputWeightArray[i][j];
                }
            }
 
            for(int i = 0; i < this.outputNodeCount; i++)
            {
                targetWeightArray[k++] = this.outputNodeBiasArray[i];
            }

            return targetWeightArray;
        }

        #endregion
        #region 가중치 배열 설정하기 - SetWeightArray(sourceWeightArray)

        /// <summary>
        /// 가중치 배열 설정하기
        /// </summary>
        /// <param name="sourceWeightArray">소스 가중치 배열</param>
        public void SetWeightArray(double[] sourceWeightArray)
        {
            int totalWeightCount = (this.inputNodeCount  * this.hiddenNodeCount) +
                                   (this.hiddenNodeCount * this.outputNodeCount) +
                                    this.hiddenNodeCount + this.outputNodeCount;
 
            if(sourceWeightArray.Length != totalWeightCount)
            {
                throw new Exception("잘못된 소스 가중치 배열 길이 입니다.");
            }
 
            int k = 0;

            for(int i = 0; i < this.inputNodeCount; i++)
            {
                for(int j = 0; j < this.hiddenNodeCount; j++)
                {
                    this.inputHiddenWeightArray[i][j] = sourceWeightArray[k++];
                }
            }

            for(int i = 0; i < this.hiddenNodeCount; i++)
            {
                this.hiddenNodeBiasArray[i] = sourceWeightArray[k++];
            }

            for(int i = 0; i < this.hiddenNodeCount; i++)
            {
                for(int j = 0; j < this.outputNodeCount; j++)
                {
                    this.hiddenOutputWeightArray[i][j] = sourceWeightArray[k++];
                }
            }

            for(int i = 0; i < this.outputNodeCount; i++)
            {
                this.outputNodeBiasArray[i] = sourceWeightArray[k++];
            }
        }

        #endregion
        #region 학습하기 - Train(trainingValueArray, maximumEpochCount, learningRate, momentum)

        /// <summary>
        /// 학습하기
        /// </summary>
        /// <param name="trainingValueArray">학습 값 배열</param>
        /// <param name="maximumEpochCount">최대 시대 카운트</param>
        /// <param name="learningRate">학습률</param>
        /// <param name="momentum">모멘텀</param>
        public void Train(double[][] trainingValueArray, int maximumEpochCount, double learningRate, double momentum)
        {
            int      epochCount          = 0;
            double[] xValueArray         = new double[this.inputNodeCount ];
            double[] targetValueArray    = new double[this.outputNodeCount];
            int      traingingValueCount = trainingValueArray.Length;
            int[]    sequenceArray       = new int[traingingValueCount];
            int      sequenceCount       = sequenceArray.Length;

            for(int i = 0; i < sequenceCount; ++i)
            {
                sequenceArray[i] = i;
            }
 
            while(epochCount < maximumEpochCount)
            {
                double mse = GetMeanSquaredError(trainingValueArray);
 
                if(mse < 0.04d)
                {
                    break;
                }
 
                Shuffle(sequenceArray);
 
                for(int i = 0; i < traingingValueCount; i++)
                {
                    int sequenceIndex = sequenceArray[i];

                    Array.Copy(trainingValueArray[sequenceIndex], xValueArray, this.inputNodeCount);

                    Array.Copy(trainingValueArray[sequenceIndex], this.inputNodeCount, targetValueArray, 0, this.outputNodeCount);

                    ComputeOutputNodeValueArray(xValueArray);

                    UpdateWeightArray(targetValueArray, learningRate, momentum);
                }
 
                epochCount++;
            }
        }

        #endregion
        #region 출력 노드 배열 계산하기 - ComputeOutputNodeValueArray(sourceInputNodeValueArray)

        /// <summary>
        /// 출력 노드 배열 계산하기
        /// </summary>
        /// <param name="sourceInputNodeValueArray">소스 입력 노드 값 배열</param>
        /// <returns>출력 노드 값 배열</returns>
        public double[] ComputeOutputNodeValueArray(double[] sourceInputNodeValueArray)
        {
            if(sourceInputNodeValueArray.Length != this.inputNodeCount)
            {
                throw new Exception("잘못된 소스 입력 노드 값 배열 길이 입니다.");
            }

            double[] hiddenNodeSummaryArray = new double[this.hiddenNodeCount];
            double[] outputNodeSummaryArray = new double[this.outputNodeCount];

            int sourceInputNodeValueCount = sourceInputNodeValueArray.Length;

            for(int i = 0; i < sourceInputNodeValueCount; i++)
            {
                this.inputNodeValueArray[i] = sourceInputNodeValueArray[i];
            }

            for(int i = 0; i < this.hiddenNodeCount; i++)
            {
                for(int j = 0; j < this.inputNodeCount; j++)
                {
                    hiddenNodeSummaryArray[i] += this.inputNodeValueArray[j] * this.inputHiddenWeightArray[j][i];
                }
            }
 
            for(int i = 0; i < this.hiddenNodeCount; i++)
            {
                hiddenNodeSummaryArray[i] += this.hiddenNodeBiasArray[i];
            }

            for(int i = 0; i < this.hiddenNodeCount; i++)
            {
                this.hiddenNodeValueArray[i] = GetHyperTangentValue(hiddenNodeSummaryArray[i]);
            }

            for(int i = 0; i < this.outputNodeCount; i++)
            {
                for(int j = 0; j < this.hiddenNodeCount; j++)
                {
                    outputNodeSummaryArray[i] += this.hiddenNodeValueArray[j] * this.hiddenOutputWeightArray[j][i];
                }
            }

            for(int i = 0; i < this.outputNodeCount; i++)
            {
                outputNodeSummaryArray[i] += this.outputNodeBiasArray[i];
            }

            double[] softMaximumArray = GetSoftMaximumArray(outputNodeSummaryArray);

            Array.Copy(softMaximumArray, this.outputNodeValueArray, softMaximumArray.Length);

            double[] targetOutputNodeValueArray = new double[this.outputNodeCount];

            Array.Copy(this.outputNodeValueArray, targetOutputNodeValueArray, targetOutputNodeValueArray.Length);

            return targetOutputNodeValueArray;
        }

        #endregion
        #region 정확도 구하기 - GetAccuracy(trainingValueArray)

        /// <summary>
        /// 정확도 구하기
        /// </summary>
        /// <param name="trainingValueArray">훈련 값 배열</param>
        /// <returns>정확도</returns>
        public double GetAccuracy(double[][] trainingValueArray)
        {
            int      correctCount       = 0;
            int      wrongCount         = 0;
            double[] xValueArray        = new double[this.inputNodeCount ];
            double[] targetValueArray   = new double[this.outputNodeCount];
            double[] yValueArray;
            int      trainingValueCount = trainingValueArray.Length;
 
            for(int i = 0; i < trainingValueCount; i++)
            {
                Array.Copy(trainingValueArray[i], xValueArray, this.inputNodeCount);

                Array.Copy(trainingValueArray[i], this.inputNodeCount, targetValueArray, 0, this.outputNodeCount);

                yValueArray = ComputeOutputNodeValueArray(xValueArray);

                int maximumIndex = GetMaximumIndex(yValueArray);

                if(targetValueArray[maximumIndex] == 1d)
                {
                    correctCount++;
                }
                else
                {
                    wrongCount++;
                }
            }

            return (correctCount * 1d) / (correctCount + wrongCount);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 매트릭스 배열 구하기 - GetMatrixArray(rowCount, columnCount)

        /// <summary>
        /// 매트릭스 배열 구하기
        /// </summary>
        /// <param name="rowCount">행 카운트</param>
        /// <param name="columnCount">컬럼 카운트</param>
        /// <returns>매트릭스 배열</returns>
        private static double[][] GetMatrixArray(int rowCount, int columnCount)
        {
            double[][] matrixArray = new double[rowCount][];

            int count = matrixArray.Length;

            for(int i = 0; i < count; i++)
            {
                matrixArray[i] = new double[columnCount];
            }

            return matrixArray;
        }

        #endregion
        #region 가중치 배열 초기화 하기 - InitializeWeightArray()

        /// <summary>
        /// 가중치 배열 초기화 하기
        /// </summary>
        private void InitializeWeightArray()
        {
            int totalWeightCount = (this.inputNodeCount  * this.hiddenNodeCount) +
                                   (this.hiddenNodeCount * this.outputNodeCount) +
                                    this.hiddenNodeCount + this.outputNodeCount;
 
            double[] initialWeightArray = new double[totalWeightCount];
            double   low                = -0.01d;
            double   high               =  0.01d;
            int      initialWeightCount = initialWeightArray.Length;

            for(int i = 0; i < initialWeightCount; i++)
            {
                initialWeightArray[i] = (high - low) * _random.NextDouble() + low;
            }
 
            SetWeightArray(initialWeightArray);
        }

        #endregion
        #region 하이퍼 탄젠트 값 구하기 - GetHyperTangentValue(x)

        /// <summary>
        /// 하이퍼 탄젠트 값 구하기
        /// </summary>
        /// <param name="x">X</param>
        /// <returns>하이퍼 탄젠트 값</returns>
        private static double GetHyperTangentValue(double x)
        {
            if(x < -20d)
            {
                return -1d;
            }
            else if(x > 20d)
            {
                return 1d;
            }
            else
            {
                return Math.Tanh(x);
            }
        }

        #endregion
        #region SOFT-MAX 배열 구하기 - GetSoftMaximumArray(sourceOutputNodeSummaryArray)

        /// <summary>
        /// SOFT-MAX 배열 구하기
        /// </summary>
        /// <param name="sourceOutputNodeSummaryArray">소스 출력 노드 합산 배열</param>
        /// <returns>SOFT-MAX 배열</returns>
        private static double[] GetSoftMaximumArray(double[] sourceOutputNodeSummaryArray)
        {
            int sourceOutputNodeSummaryCount = sourceOutputNodeSummaryArray.Length;

            double maximum = sourceOutputNodeSummaryArray[0];

            for(int i = 0; i < sourceOutputNodeSummaryCount; i++)
            {
                if(sourceOutputNodeSummaryArray[i] > maximum)
                {
                    maximum = sourceOutputNodeSummaryArray[i];
                }
            }
 
            double scale = 0d;
 
            for(int i = 0; i < sourceOutputNodeSummaryCount; ++i)
            {
                scale += Math.Exp(sourceOutputNodeSummaryArray[i] - maximum);
            }
 
            double[] targetArray = new double[sourceOutputNodeSummaryCount];
 
            for(int i = 0; i < sourceOutputNodeSummaryCount; ++i)
            {
                targetArray[i] = Math.Exp(sourceOutputNodeSummaryArray[i] - maximum) / scale;
            }
 
            return targetArray;
        }

        #endregion
        #region 가중치 배열 갱신하기 - UpdateWeightArray(targetValueArray, learningRate, momentum)

        /// <summary>
        /// 가중치 배열 갱신하기
        /// </summary>
        /// <param name="targetValueArray">타겟 값 배열</param>
        /// <param name="learningRate">학습률</param>
        /// <param name="momentum">모멘텀</param>
        private void UpdateWeightArray(double[] targetValueArray, double learningRate, double momentum)
        {
            if(targetValueArray.Length != this.outputNodeCount)
            {
                throw new Exception("타겟 값 배열 길이가 출력 노드 값 배열과 같지 않습니다.");
            }
 

            for(int i = 0; i < this.outputNodeCount; i++)
            {
                double derivative = (1 - this.outputNodeValueArray[i]) * this.outputNodeValueArray[i];
 
                this.outputNodeGradientArray[i] = derivative * (targetValueArray[i] - this.outputNodeValueArray[i]);
            }
 

            for(int i = 0; i < hiddenNodeCount; i++)
            {
                double derivative = (1 - this.hiddenNodeValueArray[i]) * (1 + this.hiddenNodeValueArray[i]);
                double summary    = 0d;
 
                for(int j = 0; j < this.outputNodeCount; j++)
                {
                    double x = this.outputNodeGradientArray[j] * this.hiddenOutputWeightArray[i][j];

                    summary += x;
                }

                this.hiddenNodeGradientArray[i] = derivative * summary;
            }
 

            for(int i = 0; i < this.inputNodeCount; i++)
            {
                for(int j = 0; j < this.hiddenNodeCount; j++)
                {
                    double delta = learningRate * this.hiddenNodeGradientArray[j] * this.inputNodeValueArray[i];

                    this.inputHiddenWeightArray[i][j] += delta;
 
                    this.inputHiddenWeightArray[i][j] += momentum * this.inputHiddenPreviousWeightDeltaArray[i][j];

                    this.inputHiddenPreviousWeightDeltaArray[i][j] = delta;
                }
            }
 

            for(int i = 0; i < this.hiddenNodeCount; i++)
            {
                double delta = learningRate * this.hiddenNodeGradientArray[i];

                this.hiddenNodeBiasArray[i] += delta;
                this.hiddenNodeBiasArray[i] += momentum * this.hiddenNodePreviousBiasDeltaArray[i];

                this.hiddenNodePreviousBiasDeltaArray[i] = delta;
            }


            for(int i = 0; i < this.hiddenNodeCount; i++)
            {
                for(int j = 0; j < this.outputNodeCount; j++)
                {
                    double delta = learningRate * this.outputNodeGradientArray[j] * this.hiddenNodeValueArray[i];

                    this.hiddenOutputWeightArray[i][j] += delta;
                    this.hiddenOutputWeightArray[i][j] += momentum * this.hiddenOutputPreviousWeightDeltaArray[i][j];

                    this.hiddenOutputPreviousWeightDeltaArray[i][j] = delta;
                }
            }
 

            for(int i = 0; i < this.outputNodeCount; i++)
            {
                double delta = learningRate * this.outputNodeGradientArray[i] * 1d;

                this.outputNodeBiasArray[i] += delta;
                this.outputNodeBiasArray[i] += momentum * this.outputNodePreviousBiasDeltaArray[i];

                this.outputNodePreviousBiasDeltaArray[i] = delta;
            } 
        }

        #endregion
        #region 섞기 - Shuffle(sequenceArray)

        /// <summary>
        /// 섞기
        /// </summary>
        /// <param name="sequenceArray">시퀀스 배열</param>
        private static void Shuffle(int[] sequenceArray)
        {
            for(int i = 0; i < sequenceArray.Length; i++)
            {
                int randomIndex = _random.Next(i, sequenceArray.Length);
                int sequence    = sequenceArray[randomIndex];
 
                sequenceArray[randomIndex] = sequenceArray[i];

                sequenceArray[i] = sequence;
            }
        }

        #endregion
        #region 평균 제곱 에러 구하기 - GetMeanSquaredError(trainingValueArray)

        /// <summary>
        /// 평균 제곱 에러 구하기
        /// </summary>
        /// <param name="trainingValueArray">훈련 값 배열</param>
        /// <returns>평균 제곱 에러</returns>
        private double GetMeanSquaredError(double[][] trainingValueArray)
        {
            double   summarySquaredError = 0d;
            double[] xValueArray         = new double[this.inputNodeCount ];
            double[] targetValueArray    = new double[this.outputNodeCount];
            int      trainingValueCount  = trainingValueArray.Length;

            for(int i = 0; i < trainingValueCount; i++)
            {
                Array.Copy(trainingValueArray[i], xValueArray, this.inputNodeCount);

                Array.Copy(trainingValueArray[i], this.inputNodeCount, targetValueArray, 0, this.outputNodeCount);

                double[] yValueArray = ComputeOutputNodeValueArray(xValueArray);

                for(int j = 0; j < this.outputNodeCount; j++)
                {
                    double error = targetValueArray[j] - yValueArray[j];

                    summarySquaredError += error * error;
                }
            }

            return summarySquaredError / trainingValueCount;
        }

        #endregion
        #region 최대 인덱스 구하기 - GetMaximumIndex(vectorArray)

        /// <summary>
        /// 최대 인덱스 구하기
        /// </summary>
        /// <param name="vectorArray">벡터 배열</param>
        /// <returns>최대 인덱스</returns>
        private static int GetMaximumIndex(double[] vectorArray)
        {
            int    maximumIndex = 0;
            double maximum      = vectorArray[0];
            int    vectorCount  = vectorArray.Length;
 
            for(int i = 0; i < vectorCount; i++)
            {
                if(vectorArray[i] > maximum)
                {
                    maximum      = vectorArray[i];
                    maximumIndex = i;
                }
            }
 
            return maximumIndex;
        }

        #endregion
    }
}